#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <errno.h>

#include "MIME_headers.h"
#include "XAM_strtok.h"
#include "strlower.h"

char *MIMEH_headerline = NULL;
int MIMEH_save_headers = 0;
int MIMEH_test_mailbox = 0;
char hexconv[256];

FILE *MIMEH_fh;


int MIMEH_set_mailbox( int level )
{
	MIMEH_test_mailbox = level;
	return level;
}

/*---------------------------------------------------
* MIME_init_hexconv()
*
 * Initialises the hex->dec conversion
*
 */
int MIMEH_init_hexconv( void ){

	hexconv['0'] = 0;
	hexconv['1'] = 1;
	hexconv['2'] = 2;
	hexconv['3'] = 3;
	hexconv['4'] = 4;
	hexconv['5'] = 5;
	hexconv['6'] = 6;
	hexconv['7'] = 7;
	hexconv['8'] = 8;
	hexconv['9'] = 9;
	hexconv['a'] = 10;
	hexconv['b'] = 11;
	hexconv['c'] = 12;
	hexconv['d'] = 13;
	hexconv['e'] = 14;
	hexconv['f'] = 15;
	hexconv['A'] = 10;
	hexconv['B'] = 11;
	hexconv['C'] = 12;
	hexconv['D'] = 13;
	hexconv['E'] = 14;
	hexconv['F'] = 15;

	return 0;
}


/*------------------------------------------------------------------------
Procedure:     MIMEH_decode_qp ID:1
Purpose:       Decodes a given line from QuotePrintable format to plain text
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIMEH_decode_qp( char *line ) {

	char c;								/* The Character to output */
	int op, ip; 						/* OutputPointer and InputPointer */
	int slen = strlen(line); /* Length of our line */

	/* Initialise our "pointers" to the start of the encoded string */
	ip=op=0;

	/* for every character in the string... */
	for (ip = 0; ip < slen; ip++){

		c = line[ip];

		/* if we have the quote-printable esc char, then lets get cracking */
		if (c == '=')
		{
			/* if we have another two chars... */
			if (ip <= (slen-2)){

				/* convert our encoded character from HEX -> decimal */
				c = (char)hexconv[(int)line[ip+1]]*16 +hexconv[(int)line[ip+2]];

				/* shuffle the pointer up two spaces */
				ip+=2;
			} /* if there were two extra chars after the ='s */

			/* if we didn't have enough characters, then  we'll make the char the
			* string terminator (such as what happens when we get a =\n
			*/
			else {
				line[ip] = '\0';
			} /* else */

		} /* if c was a encoding char */

		/* put in the new character, be it converted or not */
		line[op] = c;

		/* shuffle up the output line pointer */
		op++;
	} /* for */

	/* terminate the line */
	line[op]='\0';

	return 0;

}







/*------------------------------------------------------------------------
Procedure:     MIMEH_set_headers_save ID:1
Purpose:       Sets MIMEH's headers save file (where MIMEH will save the
headers it reads in from the mailpack)
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIMEH_set_headers_save( FILE *f )
{
	MIMEH_fh = f;
	MIMEH_save_headers = 1;
	return 1;
}




/*------------------------------------------------------------------------
Procedure:     MIMEH_read_headers ID:1
Purpose:       Reads from the stream F until it detects a From line, or a blank line
(end of headers)
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIMEH_read_headers( FILE *f )
{
	char line[1024];
	int totalsize=0;
	int linesize=0;
	int result = -1;
	int firstline = 1;
	char *tmp;
	char *fget_result = NULL;

	MIMEH_headerline = NULL;

	while ((fget_result=fgets(line,1024, f)))
	{

		if ((line[0] != '\n')&&(line[0] != '\r'))
		{


			if ((MIMEH_test_mailbox)&&(firstline))
			{
				if ((XAM_strncasecmp(line,"From ",5)==0)||(XAM_strncasecmp(&(line[1]),"From ",5)==0))
				{
					MIMEH_headerline = malloc(strlen(line)+1);
					strcpy(MIMEH_headerline, line);
					return _MIMEH_FOUND_FROM;
				}
			}


			linesize = strlen(line);
			totalsize += linesize;
			tmp = realloc(MIMEH_headerline, totalsize+1);

			if (!tmp)
			{
				syslog(1,"MIMEH_read_headers(): cannot allocate %d bytes ",totalsize);
				free(MIMEH_headerline);
				MIMEH_headerline = NULL;
				return -1;
			}

			if (!MIMEH_headerline)
			{
				MIMEH_headerline = tmp;
				strcpy(MIMEH_headerline, line);
			}
			else
			    {
				MIMEH_headerline = tmp;
				strcat(MIMEH_headerline,line);
			}

		} // If line doesn't start with a \n or \r

		else
		{
			if ((MIMEH_save_headers)&&(MIMEH_headerline))
			{
				fprintf(MIMEH_fh,"%s",MIMEH_headerline);
			}
			result = 1;
			break;
		}

		firstline = 0;
	}

	if (!fget_result)
	{
		result = -1;
	}

	return result;
}



/*------------------------------------------------------------------------
 Procedure:     MIMEH_display_info ID:1
 Purpose:       DEBUGGING - Displays the values of the hinfo structure to
                stderr
 Input:
 Output:
 Errors:
------------------------------------------------------------------------*/
int MIMEH_display_info( struct _header_info *hinfo )
{
	if (hinfo)
	{
		fprintf(stdout,"\
Content Type = %d\n\
Boundary = %s\n\
Filename = %s\n\
    name = %s\n\
Encoding = %d\n\
Disposit = %d\n\
"\
,hinfo->content_type\
,hinfo->boundary\
,hinfo->filename\
,hinfo->name\
,hinfo->content_transfer_encoding\
,hinfo->content_disposition);
		fflush(stdout);
	}
	return 1;
}


int MIMEH_parse_filename( char *dest, char *raw )
{
	char *p, *q, *n;
	int qp_encode = 0;
	int quoted = 0;
	char tmp[1024];

	n = strpbrk(raw,"\n\r;");
	if (n) *n = '\0';

	strncpy(tmp,raw,1023);
	p = tmp;

	//fprintf(stdout,"Parsefilename (RAW) : %s\n",p); fflush(stdout);

	if (XAM_strncasecmp(p,"\"3d",3)==0)
	{
		qp_encode = 1;
		quoted=1;
		p+=3;
	}
	else
		if (XAM_strncasecmp(p,"3d\"",3)==0)
		{
			qp_encode = 1;
			quoted = 1;
			p+=3;
		}
		else
			if (XAM_strncasecmp(p,"3d",2)==0)
			{
				qp_encode = 1;
				quoted=0;
				p+=2;
			}
			else
				if ((*p) == '\"')
				{
					qp_encode = 0;
					quoted = 1;
					p++;
				}



	//	fprintf(stdout,"P1 = %s\n",p);

	if ((qp_encode)&&(strstr(p,"=OA="))) qp_encode = 1;
	else qp_encode = 0;

	//		fprintf(stdout,"P2 = %s\n",p);

	sprintf(dest,"%s",p);

	//		fprintf(stdout,"Dest1 = %s\n",dest);
		if (qp_encode) MIMEH_decode_qp(dest);
	//		fprintf(stdout,"Dest2 = %s\n",dest);

	if (quoted)
	{
		//		fprintf(stdout,"Dest3 = %s\n",dest);
			q = strrchr(dest,'\"');
		//		fprintf(stdout,"Dest4 = %s\n",dest);
			if (q) *q = '\0';
		//		fprintf(stdout,"Dest5 = %s\n",dest);
			}

	//	fprintf(stdout,"\nOutput filename = %s\n",dest);

	if (n) *n = '\n';

	return 1;

}

/*------------------------------------------------------------------------
Procedure:     MIMEH_parse_headers ID:1
Purpose:       Reads an open filestream in and parses the headers for it.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int   MIMEH_parse_headers( FILE *f, struct _header_info *hinfo )
{
	char *safeh, *h, *hl, *safehl;
	char *p,*q, c, *pre_p;
	char *lba;
	char *line_end;
	char *next_angle;
	int result;

	MIMEH_init_hexconv();

	result = MIMEH_read_headers(f);

	if (result == _MIMEH_FOUND_FROM)
	{
		if (MIMEH_headerline)
		{
			free(MIMEH_headerline);
			MIMEH_headerline = NULL;
		}

		return _MIMEH_FOUND_FROM;
	}



	safeh = h = MIMEH_headerline;


	if (result == -1)
	{
		return -1;
	}

	if (!MIMEH_headerline)
	{
		return 1;
	}


	// Duplicate string
	hl = malloc(sizeof(char) *(strlen(h)+1));
	strcpy(hl,h);
	strlower(hl);
	safehl = hl;
//		fprintf(stdout,"SCANNING THROUGH THE HEADERS----------------\n%s\n",hl);
	// Now lets go looking...





	// CONTENT TYPE -------------------------------
	// CONTENT TYPE -------------------------------
	// CONTENT TYPE -------------------------------

	p = strstr(hl,"content-type");
	if (p)
	{
		p = p +strlen("content-type");
		q = strpbrk(p,"\n\r;");
		if (q)
		{
			c = *q;
			*q = '\0';
			if (strstr(p,"multipart/")) hinfo->content_type = _CTYPE_MULTIPART;
			else
			    if (strstr(p,"text/")) hinfo->content_type = _CTYPE_TEXT;
			else
			    if (strstr(p,"message/rfc822")) hinfo->content_type = _CTYPE_RFC822;
			else
			    if (strstr(p,"/octet-stream")) hinfo->content_type = _CTYPE_OCTECT;
			else
				if (strstr(p,"/ms-tnef")) hinfo->content_type = _CTYPE_TNEF;
			else hinfo->content_type = _CTYPE_UNKNOWN;
			*q = c;
		}
		else
		{
			syslog(1,"MIMEH_parse_headers(): Cannot locate end of Content-Type specifier!");
		}
	}
	else
	{
		// If we cant find a content-type: specifier in these headers
		// then we've really got a bundle on our hands... either we
		// overshot on reading something somewhere, or we have some
		// crazed input.  Either way, return 1
		return 1;
	}




	// Seeking out boundaries is not quite as simple as it seems.
	// we can "initially" think that we can just do a "strstr", but
	// what if there is a filename="someboundary=OAxxx.gif" then
	// we're in trouble!
	//
	lba = hl;
	p = strstr(lba, "boundary=");
	while ( (lba != NULL) && (hinfo->boundary_located==0)  && ( p != NULL) )
	{
		if (p == NULL) break;

//		fprintf(stdout,"TRYING boundary : %s",p);

		// Setup a pointer to the char just before this boundary= location
		if (p > hl)
		{
			pre_p = p -1;
		}

		//		fprintf(stdout,"Testing boundary at: %s",p);
		p = p +strlen("boundary=");
		p = h +(p -hl);

		// If we have a quoted boundary, shift one char to the right
		if ((*p) == '\"') p++;

		// Try to find the end of the boundary by looking for ", \n, \r \t or ;
		q = strpbrk(p,"\"\n\r\t;");

		// If we did find something for the end of the boundary
		if (q)
		{
			// get the character which we stopped on
			c = *q;

			// turn the location into a string terminator
			*q = '\0';

			// If the first char of the boundary isn't a normal alphanumeric char
			if (!(isprint(*p)))
			{
				// Put the original character back
				*q = c;

				// Move our start point (for detecting boundaries in the headers)
				// alone one more char...
				lba++;

				// Try find a new boundary...
				p = strstr(lba,"boundary=");
				continue;
			}
			else
			    {
				// if the character was valid, then we have detected a boundary
				// specifier!.  Copy this to the hinfo structure, and set the
				// "located" flag
				sprintf(hinfo->boundary,"%s",p);
				hinfo->boundary_located = 1;
			}

			// Set our string terminator location back to its original character
			// (if we dont, then any further string operations wont see beyond
			// it)
			*q = c;
		}
		else
		{
			syslog(1,"MIMEH_parse_headers(): Cannot locate end of boundary specifier!");
		}
	} // while


	hl = safehl;







	// ATTACHMENT FILENAME ----------------------------

	// ATTACHMENT FILENAME ----------------------------

	// ATTACHMENT FILENAME ----------------------------

	p = strstr(hl,"filename=");
	if (p)
	{
		p = p +strlen("filename=");
		p = h +(p -hl);
		MIMEH_parse_filename(hinfo->filename, p);
	}


	// ATTACHMENT -NAME ----------------------------

	// ATTACHMENT -NAME ----------------------------

	// ATTACHMENT -NAME ----------------------------

	if (hinfo->filename[0] == '\0')
	{
		p = strstr(hl,"name=");
		if ( (p) && (isspace(*(p-1))) )
		{
			/* Because HTML in MIME is so common these days, it's not
			unusual to get a <META .... name="..."> tag in the headers
			hence, it's a good idea to text for the trailing > (which
			is not a legal filechar anyhow!) and REJECT if it contains
			it */
			next_angle = line_end = NULL;
			next_angle=strchr(p,'>');
			if (next_angle) line_end=strpbrk(p,"\n\r\";");
			if ( (!next_angle) || (line_end && (line_end < next_angle)) )
			{
				p = p +strlen("name=");
				p = h +(p -hl);
				MIMEH_parse_filename(hinfo->name, p);
			}
		}
	}


	if ((hinfo->filename[0] == '\0')&&(hinfo->name[0] != '\0'))
		strncpy(hinfo->filename, hinfo->name, sizeof(hinfo->filename));




	// CONTENT TRANSFER ENCODING ---------------------

	// CONTENT TRANSFER ENCODING ---------------------

	// CONTENT TRANSFER ENCODING ---------------------

	p = strstr(hl,"content-transfer-encoding");
	if (p)
	{
		p = p +strlen("content-transfer-encoding");
		q = strpbrk(p,"\n\r;");
		if (q)
		{
			c = *q;
			*q = '\0';
			if (strstr(p,"base64")) hinfo->content_transfer_encoding = _CTRANS_ENCODING_B64;
			else
			    if (strstr(p,"7bit")) hinfo->content_transfer_encoding = _CTRANS_ENCODING_7BIT;
			else
			    if (strstr(p,"8bit")) hinfo->content_transfer_encoding = _CTRANS_ENCODING_8BIT;
			else
			    if (strstr(p,"quoted-printable")) hinfo->content_transfer_encoding = _CTRANS_ENCODING_QP;
			else hinfo->content_transfer_encoding = _CTRANS_ENCODING_UNKNOWN;
			*q = c;
		}
		else
		{
			syslog(1,"MIMEH_parse_headers(): Cannot locate end of Content-Transfer-Encoding specifier!");
		}
	}







	// CONTENT DISPOSITION ------------------------------

	// CONTENT DISPOSITION ------------------------------

	// CONTENT DISPOSITION ------------------------------

	p = strstr(hl,"content-disposition");
	if (p)
	{
		p = p +strlen("content-disposition");
		q = strpbrk(p,"\n\r;");
		if (q)
		{
			c = *q;
			*q = '\0';
			if (strstr(p,"inline")) hinfo->content_disposition = _CDISPOSITION_INLINE;
			else
			    if (strstr(p,"attachment")) hinfo->content_disposition = _CDISPOSITION_ATTACHMENT;
			else hinfo->content_type = _CDISPOSITION_UNKNOWN;
			*q = c;
		}
		else
		{
			syslog(1,"MIMEH_parse_headers(): Cannot locate end of Content-Disposition specifier!");
		}
	}


	if (MIMEH_headerline)
	{
		free(MIMEH_headerline);
		MIMEH_headerline = NULL;
	}

	return 1;
}

