/*------------------------------------------------------
*
 * MIME.C
*
 * This file is the core component of the ripMIME
* and other packages
*
 * Sun Jan 28 16:16:12 EST 2001 worker <worker@xamime>
* 	Prerelease stage 1 completed.
* Sat Jan 13 21:05:51 EST 2001 worker <worker@xamime>
* 	Going into PreRelease stage 1.
*
 * Sat Nov 25 16:25:06 SAST 2000 worker <worker@xamime>
* 	Add some more comments, as more people are reading this
* 	c code ;)
*
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include "XAM_strtok.h"
#include "strlower.h"
#include "ffget.h"
#include "mime.h"
#include "MIME_headers.h"
#include "tnef/tnef_api.h"
#define _SL 1

#define _ENC_UNKNOWN 0
#define _ENC_BASE64 1
#define _ENC_PLAINTEXT 2
#define _ENC_QUOTED 3
#define _ENC_EMBEDDED 4
#define _ENC_NOFILE -1
#define _MIME_CHARS_PER_LINE 32
#define _MIME_MAX_CHARS_PER_LINE 76
#define _RECURSION_LEVEL_MAX 20

#define _BOUNDARY_CRASH 2

/* single-character decode */
#define UUDEC(c)  (((c) - ' ') & 077)


/*----Structures-----*/
struct _MIME_info {
	int lastlinewasboundary;	/* Last line we read in had the boundary in it */
	int lastlinewasfrom;		 /* last line we read in had the From: line proceeded by a blank */
	int lastlinewasblank;		/* (part of lastlinewasfrom) */
	int lastencoding;			 /* the encoding specifier last read in */
	int boundarylen;			  /* Length of the boundary */
	char *inputok;				 /* Indicator of file reading sanity */
	char boundary[513]; 		 /* The boundary */
	char filename[513];		  /* Filename of current attachment */
};


/* our base 64 decoder table */
unsigned char b64[256];

char OK[]="OKAY";

/* HexConversion table, rather than doing a atoi (with base 16) */
int hexconv[256];

/* How many attachments have we produced */
int filecount = 0;

/* If the attachment has no name, then we will need it to have one (else how are you going store it on your system ? */
char blankfileprefix[128]="textfile";

/* Verbosity - talk as we walk through the MIMEpack to the shell */
int _verbosity = 0;

/* Our logging options */
int syslogging = 0;
int stderrlogging = 1;

/* filename options */
int _unique_names = 0;
int _no_paranoid = 0;
int _rename_method = _MIME_RENAME_METHOD_INFIX;

/* The name of the file which will contain all the non-attachment
* lines from the MIMEpack
*/
char headersname[1024]="_headers_";

/* The variable that indicates (if not zero) that we are to dump our
* non-attachment lines to the file "headersname"
*/
int _dump_headers = 0;


/* Attachment count - how many attachment with FILENAMES */
int _attachment_count = 0;

/* Current line read number */
int _current_line = 0;

/* if we dont want nameless-files, this is non-zero */
int _no_nameless = 0;

/* Are we going to handle mailbox format */
int _mailbox_format = 0;

/* File pointer for the headers output */
FILE *headers;

char _tmpdir[1024];

struct _tbsnode {
	char *boundary;
	struct _tbsnode *next;
};

struct _tbsnode *boundarystack = NULL;
char boundarystacksafe[1024];

/*----------------------------------------------------------
* MIME_BS_push()
*
 * Push a boundary onto the stack */
int MIME_BS_push( char *boundary )
{

	struct _tbsnode *node = malloc(sizeof(struct _tbsnode));


	if (node)
	{
		node->next = boundarystack;
		boundarystack = node;
		boundarystack->boundary = strdup(boundary);
	}
	else
	{
		syslog(_SL,"MIME_BS_push(): Cannot allocate memory for PUSH(), %s",strerror(errno));
	}

	return 0;
}


/*----------------------------------------------------------
* MIME_BS_pop()
*
 * pop's the top boundar off */
char *MIME_BS_pop( void )
{

	struct _tbsnode *node = boundarystack;
	if (boundarystack)
		{
		boundarystack = boundarystack->next;
		strcpy(boundarystacksafe, node->boundary);
		free(node);
	}

	return boundarystacksafe;
}

/*----------------------------------------------------------
* MIME_BS_top()
*
 * returns the top item in the stack, without popping off */
char *MIME_BS_top( void )
{

	if (boundarystack)
		{
		return boundarystack->boundary;
	}
	else return NULL;
}

/*----------------------------------------------------------
* MIME_BS_cmp()
*
 */
int MIME_BS_cmp( char *boundary )
{

	char *top, *spot;
	int spin=1;
	struct _tbsnode *node=boundarystack;
	struct _tbsnode *nodetmp=NULL, *nodedel=NULL;

	while((node)&&(spin))
	{
		top = node->boundary;
		spot = strstr(boundary,top);
		if (spot) spin = 0;
		else node = node->next;
	}

	// If we have a hit on the matching, then, according
	// to nested MIME rules, we must "remove" any previous
	// boundaries
	//
	if (spin == 0)
	{
		if (node != boundarystack)
		{
			nodetmp = boundarystack;
			while ((nodetmp)&&(nodetmp != node))
			{
				nodedel = nodetmp;
				nodetmp = nodetmp->next;
				free(nodedel);
			}
			boundarystack = node;
		}
		return 1;
	}

	else return 0;
}


/*----------------------------------------------------------
* MIME_set_tmpdir()
*
 */
int MIME_set_tmpdir( char *setto )
{

	sprintf(_tmpdir,"%s",setto);

	return 0;
}



/*-------------------------------------------------------
* MIME_set_blankfileprefix
*
* Sets the prefix for blank files
*
 * the blank file prefix is used when (typ) we have text
* attachments, ie, when there is both a HTML and a
* plain-text version of the message in the same email.
*
 */
int MIME_set_blankfileprefix( char *prefix )
{

	/* copy over the prefix name, ensure that we
	* dont break buffers here */
	strncpy(blankfileprefix,prefix,127);

	return 0;
}


/*------------------------------------------------------------------------
Procedure:     MIME_get_blankfileprefix ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
char *MIME_get_blankfileprefix( void )
{
	return blankfileprefix;
}


/*-------------------------------------------------------
* MIME_set_verbosity
*
* By default, MIME reports nothing as its working
* Setting the verbosity level > 0 means that it'll
* report things like the name of the files it's
* writing/extracting.
*
 */
int MIME_set_verbosity( int level )
{

	_verbosity = level;

	return 0;
}


/*-------------------------------------------------------
* MIME_set_dumpheaders
*
 * By default MIME wont dump the headers to a text file
* but at times this is useful esp for checking
* for new styles of viruses like the KAK.worm
*
 * Anything > 0 will make the headers be saved
*
 */
int MIME_set_dumpheaders( int level )
{

	_dump_headers = level;

	return 0;
}


/*------------------------------------------------------
* MIME_set_headersname
*
* by default, the headers would be dropped to a file
* called '_headers_'.  Using this call we can make
* it pretty much anything we like
*/
int MIME_set_headersname( char *fname )
{

	strncpy(headersname, fname, 1023);

	return 0;
}


/*-------------------------------------------------------
* MIME_set_syslogging
*
 * By default, syslogging is off, setting the "level" to a value above zero(0)
* will make MIME log to the system
*/
int MIME_set_syslogging( int level )
{

	syslogging = level;

	return 0;
}


/*-------------------------------------------------------
* MIME_set_stderrlogging
*
 * By default logging is done via stderr, set this to zero (0) to turn off
*/
int MIME_set_stderrlogging(int level)
{

	stderrlogging = level;

	return 0;
}



/*------------------------------------------------------------------------
Procedure:     MIME_set_no_nameless ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIME_set_no_nameless( int level )
{

	_no_nameless = level;

	return 0;
}

/*------------------------------------------------------------------------
Procedure:     MIME_set_uniquenames ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIME_set_uniquenames( int level )
{
	_unique_names = level;

	return 0;
}


/*------------------------------------------------------------------------
 Procedure:     MIME_set_noparanoid ID:1
 Purpose:       If set, will prevent MIME from clobbering what it considers
                to be non-safe characters in the file name.
 Input:
 Output:
 Errors:
------------------------------------------------------------------------*/
int MIME_set_noparanoid( int level )
{
	_no_paranoid = level;
	return 0;
}


/*------------------------------------------------------------------------
 Procedure:     MIME_set_mailboxformat ID:1
 Purpose:       If sets the value for the _mailboxformat variable
                in MIME, this indicates to functions later on
                that they should be aware of possible mailbox
                format specifiers.
 Input:
 Output:
 Errors:
------------------------------------------------------------------------*/
int MIME_set_mailboxformat( int level )
{
	_mailbox_format = level;
	MIMEH_set_mailbox( level );
	return 0;
}


/*------------------------------------------------------------------------
Procedure:     MIME_set_renamemethod ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIME_set_renamemethod( int method )
{
	if (( method >= _MIME_RENAME_METHOD_PREFIX ) && ( method <= _MIME_RENAME_METHOD_POSTFIX ))
	{
		_rename_method = method;
	}
	else
	    {
		//		#FIXME - make me report correct values to stderr or syslog
		return -1;
	}

	return 0;
}

/*------------------------------------------------------------------------
Procedure:     MIME_get_attachment_count ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIME_get_attachment_count( void )
{

	return _attachment_count;
}


/*-------------------------------------------------------
* MIME_tmpnam()
*
 * Creates a temporary filename
*
 * NOTE - P_tmpdir comes from stdio.h, see 'man tmpfile'
*/
int MIME_tmpnam( char *_temp_fname )
{
	int i;
	char sequence[12];
	FILE *f;

	do {
		for (i=0; i <11; i++) sequence[i] = rand()%(122 -97) +97;
		sequence[11] = '\0';
		sprintf(_temp_fname,"%s/%s_ripmime_%d",_tmpdir, sequence, getpid());
	}
	while ((f=fopen(_temp_fname,"r")));

	return 0;
}










/*------------------------------------------------------------------------
Procedure:     MIME_test_uniquename ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIME_test_uniquename( char *path, char *fname, int method )
{

	struct stat buf;

	char newname[1024];
	char *scratch = malloc((strlen(fname)+2) *sizeof(char));
	char *frontname, *extention;

	int cleared = 0;
	int count = 1;

	if (!scratch)
	{
		return -1;
	}

	frontname = extention = NULL;  // shuts the compiler up

	if (method == _MIME_RENAME_METHOD_INFIX)
	{
		strncpy(scratch,fname,strlen(fname)+1);
		frontname = scratch;
		extention = strrchr(scratch,'.');
		if (extention)
		{
			*extention = '\0';
			extention++;
		}
		else
		    {
			method = _MIME_RENAME_METHOD_POSTFIX;
		}
	}

	sprintf(newname,"%s/%s",path,fname);

	while (!cleared)
	{
		if ((stat(newname, &buf) == -1))
		{
			cleared++;
		}
		else
		    {
			if (method == _MIME_RENAME_METHOD_PREFIX)
			{
				sprintf(newname,"%s/%d_%s",path,count,fname);
			}
			else
			    if (method == _MIME_RENAME_METHOD_INFIX)
			{
				sprintf(newname,"%s/%s_%d.%s",path,frontname,count,extention);
			}
			else
			    if (method == _MIME_RENAME_METHOD_POSTFIX)
			{
				sprintf(newname,"%s/%s_%d",path,fname,count);
			}
			count++;
		}
	}

	if (count > 1)
	{
		frontname = strrchr(newname,'/');
		if (frontname) frontname++;
		else frontname = newname;

		strncpy(fname, frontname, 1024);
		if (scratch) free(scratch);
	}

	return 0;
}



/*-------------------------------------------------------
* MIME_base64_init
*
 * Prepare the base64 decoding array
*
 */
int MIME_base64_init( void )
{

	int i;

	/* preset every encodment to 0x80 */
	for (i = 0; i < 255; i++) b64[i] = 0x80;

	/* put in the capital letters */
	for (i = 'A'; i <= 'Z'; i++) b64[i] = 0 + (i - 'A');

	/* put in the lowere case letters */
	for (i = 'a'; i <= 'z'; i++) b64[i] = 26 + (i - 'a');

	/* the digits */
	for (i = '0'; i <= '9'; i++) b64[i] = 52 + (i - '0');

	/* and our special chars */
	b64['+'] = 62;
	b64['/'] = 63;
	b64['='] = 0;

	return 0;
}





/*-------------------------------------------------------------------*
 * MIME_clean_MIME_filename -
* 	Cleans up the file name, ie, strips it of unrequired
*	 inverted comma's etc.
*/
int MIME_clean_MIME_filename( char *fname )
{
	int fnl = strlen(fname); /* FileNameLength */
	int i;
	char *fnp, *iso, fname_copy[1024];
	char tmp[1024];
	char *p = strrchr(fname,'/');
	struct _txstrtok xst;

	/* scan out any directory separators */
	if (p)
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}
	else if ( (p = strrchr(fname,'\\')))
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}



	if ((fnl > 2)&&(!strchr(fname,' ')))
	{  /* if there's no spaces in our MIME_filename */

		if (strstr(fname,"=?"))
		{ /* we may have an ISO filename */
			fnp = fname;
			iso = XAM_strtok(&xst,fname,"?");
			/*iso = strtok(fname,"?");  iso encoding prefix */
			if (iso) iso = XAM_strtok(&xst, NULL,"?"); /* The leading = */
			if (iso) iso = XAM_strtok(&xst, NULL,"?"); /* unknown singe char */
			if (iso) iso = XAM_strtok(&xst, NULL,"?"); /* filename! */
			if (iso) {
				MIME_decode_text_line(iso);
				sprintf(fname_copy,"%s",iso);
				sprintf(fname,"%s",fname_copy);
			}
		}
		/* if the MIME_filename starts and ends with "'s */
		if ((fname[0] == '\"') && (fname[fnl-1] == '\"'))
		{

			/* reduce the file namelength by two*/
			fnl=-2;

			/* shuffle the MIME_filename chars down */
			for (i = 0; i < fnl;) fname[i] = fname[++i];

			/* terminate the string */
			fname[fnl] = '\0';
		} /* if */
	} /* if */

	return 0;
}


/*-------------------------------------------------------------------*
 * quick_clean_filename -
* 	Cleans up filename by replacing non alfanum chars with '_'
*/
void quick_clean_filename( char *fname )
{
	char tmp[1024];
	char *p = strrchr(fname,'/');

	/* scan out any directory separators */
	if (p)
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}
	else if ( (p = strrchr(fname,'\\')))
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}


	while (*fname)
	{
		if (_no_paranoid == 0)
		{
			if( !isalnum(*fname) && (*fname != '.') ) *fname='_';
		}
		else
		    {
			if( (*fname < ' ')||(*fname > '~') ) *fname='_';
		}
		fname++;
	}
}




/*-------------------------------------------------------
* MIME_getchar_start
*
 * This routine is actually a copy of MIME_getchar()
* except that it'll return ONLY a valid MIME-encoded
* char, rather than also a \n
*
 */

int MIME_getchar_start( FILE *f )
{

	int c;

	/* loop for eternity, as we're "returning out" */
	while (1)
	{

		/* get a single char from file */
		c = FFGET_fgetc();

		/* if that char is an EOF, or the char is something beyond
		* ASCII 32 (space) then return */
		if ((c == EOF) || (c > ' '))
		{
			return c;
		}

	} /* while */

	/* Although we shouldn't ever actually get here, we best put it in anyhow */
	return EOF;
}



/*-------------------------------------------------------
* MIME_getchar()
*
 * get the next char from the MIME-encoded-stream
*
 */

int MIME_getchar( FILE *f )
{

	int c;

	/* loop for eternity, as we're "returning out" */
	while (1)
	{
		/* get a single char from file */
		c = fgetc(f);
		/* if that char is an EOF, or the char is something beyond
		* ASCII 32 (space) then return */
		if ((c == EOF) || (c > ' ')||(c == '\n'))
		{
			return c;
		}

	} /* While */

	/* Though we shouldn't get this far, we should put in something just incase! */
	return EOF;
}





/*----------------------------------------------------------------------
* MIME_insert_X_header
*
* Inserts a specified X-header at the end of other X-headers.
*/
int MIME_insert_Xheader( char *fname, char *xheader)
{

	/* Tempfile tmpfile() fix contributed by David DeMaagd - 29/01/2001 */
	char line[1024], lline[1024];
	char tpn[1024];
	int xon = 0;
	int returnvalue = 0;
	FILE *fo;
	FILE *f;

	/* Get temp name */
	sprintf(tpn,"%s.",fname);

	fo = fopen(tpn,"w");
	if (!fo)
	{
		if (syslogging > 0) syslog(_SL,"Error: Cannot open temporary file for writing (%s) : %s",tpn, strerror(errno));
		if (stderrlogging > 0) fprintf(stderr,"Error: Cannot open temporary file for writing");
		returnvalue = 1;
	}
	f = fopen(fname,"r");
	if (!f)
	{
		if (syslogging > 0) syslog(_SL,"Error: Cannot open Mailpack file for reading (%s) : %s",fname, strerror(errno));
		if (stderrlogging > 0) fprintf(stderr,"Error: Cannot open Mailpack file for reading, %s",fname);
		returnvalue = 2;
	}

	if ( returnvalue == 0 )
	{
		while (fgets(line,1023,f))
		{
			_current_line++;
			strncpy(lline,line,1023);
			strlower(lline);

			if ((xon==0)&&(strstr(lline,"content-type:")))
			{
				fprintf(fo,"%s\n",xheader);
				xon=1;
			}

			fprintf(fo,"%s",line);
		}

		/* close files */
		fclose(fo);
		fclose(f);

		/* remove the old file */
		if (remove(fname) == -1)
		{
			syslog(_SL,"Error: xInsertHeader - REMOVE, %s",strerror(errno));
		}
		else
		{
			/* and rename */
			if (rename(tpn,fname) == -1)
			{
				syslog(_SL,"Error: xInsertHeader - RENAME, %s",strerror(errno));
			} /*if rename*/
		} /* else */

	} /* if */

	return returnvalue;
}







/*-------------------------------------------------
* MIME_decode_text_line
*
 * This function just decodes a single print-quotable line into
* it's normal/original format
*/
int MIME_decode_text_line( char *line )
{

	char c;								/* The Character to output */
	int op, ip; 						/* OutputPointer and InputPointer */
	int slen = strlen(line); /* Length of our line */

	/* Initialise our "pointers" to the start of the encoded string */
	ip=op=0;

	/* for every character in the string... */
	for (ip = 0; ip < slen; ip++)
	{

		c = line[ip];

		/* if we have the quote-printable esc char, then lets get cracking */
		if (c == '=')
		{

			/* if we have another two chars... */
			if (ip <= (slen-2))
			{

				/* convert our encoded character from HEX -> decimal */
				c = (char)hexconv[(int)line[ip+1]]*16 +hexconv[(int)line[ip+2]];

				/* shuffle the pointer up two spaces */
				ip+=2;
			} /* if there were two extra chars after the ='s */

			/* if we didn't have enough characters, then  we'll make the char the
			* string terminator (such as what happens when we get a =\n
			*/
			else
			{
				line[ip] = '\0';
			} /* else */

		} /* if c was a encoding char */

		/* put in the new character, be it converted or not */
		line[op] = c;

		/* shuffle up the output line pointer */
		op++;
	} /* for */

	/* terminate the line */
	line[op]='\0';

	return 0;

}




/*------------------------------------------------------------------------
 Procedure:     MIME_decode_TNEF ID:1
 Purpose:       Decodes TNEF encoded attachments
 Input:
 Output:
 Errors:
------------------------------------------------------------------------*/
int MIME_decode_TNEF( FILE *f, char *unpackdir, struct _header_info *hinfo, int keep )
{
	int result=0;
	char arg0[]="ripmime";
	char arg1[]="-s";
	char *args[3];

	args[0] = arg0;
 	args[1] = arg1;
	args[2] = hinfo->filename;

	TNEF_main(3, args);

	return result;
}




/*------------------------------------------------------------------------
 Procedure:     MIME_decode_text ID:1
 Purpose:       Decodes an input stream into a text file.
 Input:         unpackdir : directory where to place new text file
                hinfo : struct containing information from the last parsed headers
                keep : if set, retain the file
 Output:
 Errors:
------------------------------------------------------------------------*/
int MIME_decode_text( FILE *f, char *unpackdir, struct _header_info *hinfo, int keep )
{

	FILE *of; 							// output file
	int linecount = 0;  				// The number of lines
	char fullfilename[1024]=""; 	// Filename of the output file
	char line[1024]; 					// The input lines from the file we're decoding
	char *get_result = &line[0];
	int lastlinewasboundary = 0;
	int result = 1;

	sprintf(fullfilename,"%s/%s",unpackdir,hinfo->filename);

	if (!f)
	{
		if (syslogging > 0) syslog(_SL,"Error: ripMIME print-quotable input stream broken.");
		if (stderrlogging > 0) fprintf(stderr,"Error: ripMIME print-quotable input stream broken.");
		//exit(_EXITERR_PRINT_QUOTABLE_INPUT_NOT_OPEN);
		return -1;
	}

	// if our FILE stream is open
	if (f)
	{

		of = fopen(fullfilename,"w");
		if (!of)
		{
			if (syslogging > 0) syslog(_SL,"Error: cannot open %s for writing",fullfilename);
			if (stderrlogging > 0) fprintf(stderr,"Error: cannot open %s for writing",fullfilename);
			//			exit(_EXITERR_PRINT_QUOTABLE_OUTPUT_NOT_OPEN);
			return -1;
		}

		while ((get_result = fgets(line,1023,f))&&(of))
		{

			if (_mailbox_format > 0)
			{
				if ((line[0] == '\n')||(line[0] == '\r'))
				{
					get_result = fgets(line,1023,f);
					if (get_result)
					{
						if ((XAM_strncasecmp(line, "from ",5)==0)||(XAM_strncasecmp(&(line[1]),"from ",5)==0))
						{
							result = _MIMEH_FOUND_FROM;
							break;
						}
					} // if (get_result)
					else break;

				} // if blank line

			} // if mailbox format


			if ((hinfo->boundary[0] != '\0')&&(MIME_BS_cmp(line)))
			{
//				fprintf(stderr,"Strike on boundary..line:%d:%s",linecount,line);
				lastlinewasboundary = 1;
				result = 1;
				break;
			}

			/* If we're dealing with a mailbox format mail file, then we
			need to first check to see if the line is a blank one, if so
			we flag this, then, the next time we cycle around, if we
			have a "From" or ">From" line, we can exit, citing that we
			have gotten to the end of the last text.
			*/

			if (lastlinewasboundary == 0)
			{
				if (hinfo->content_transfer_encoding == _CTRANS_ENCODING_QP)  MIME_decode_text_line(line);
				fprintf(of,"%s",line);
			}
			linecount++;

		} // while





		// if the file is still safely open
		if (of)
		{

			// close it
			fclose(of);
			// if we wrote nothing, then trash the file
			if ((keep == 0)||(linecount == 0))
			{
				unlink(fullfilename);
			}
		} // if file still safely open
	} // if main input file stream was open


	// If our input from the file was invalid due to EOF or other
	// then we return a -1 code to indicate that the end has
	// occured.
	//
	if (!get_result) result = -1;


	return result;
}





/*-------------------------------------------------------------------*
 * MIME_decode64 -
* 	Decodes base64 type streams
*
 * This routine is very very very important, it's the key to ensuring
* we get our attachments out of the email file without trauma!
*
 * NOTE - this has been -slightly altered- in order to make provision
* of the fact that the attachment may end BEFORE the EOF is received
* as is the case with multiple attachments in email.  Hence, we
* now have to detect the start character of the "boundary" marker
*
 * I may consider testing the 1st n' chars of the boundary marker
* just incase it's not always a hypen '-'.
*
 */
int MIME_decode_64( FILE *f, char *unpackdir, struct _header_info *hinfo )
{

	int i;
	int t_length = 0; /* the length of the current line we're decoding */
	int p_length = -1; /* the length of the PREVIOUS line we were decoding */
	int cr_count = 0; /* the number of consecutive \n's we've read in, used to detect End of B64 enc */
	int stopcount = 0; /* How many stop (=) characters we've read in */
	int eom_reached = 0; /* flag to say that we've reached the End-Of-MIME encoding. */
	int status = 1; /* Overall status of decoding operation */
	int c; /* a single char as retrieved using MIME_get_char() */
	int char_count = 0; /* How many chars have been received */
	int boundary_crash = 0; /* if we crash into a boundary, set this */
	long int bytecount=0; /* The total file decoded size */
	char output[3]; /* The 4->3 byte output array */
	char input[4]; /* The 4->3 byte input array */
	char fullMIME_filename[1024]=""; /* Full Filename of output file */

	FILE *of; /* output file pointer */

	/* setup our input data stream */
	FFGET_setstream(f);

	/* generate the MIME_filename, and open it up... */
	if (_unique_names) MIME_test_uniquename( unpackdir, hinfo->filename, _rename_method );
	sprintf(fullMIME_filename,"%s/%s",unpackdir,hinfo->filename);
	of = fopen(fullMIME_filename,"wb");


	/* if we were unable to open the output file, then we better log an error and drop out */
	if (!of)
	{
		if (syslogging > 0) syslog(_SL,"Error: Cannot open output file %s for BASE64 decoding.",fullMIME_filename);
		if (stderrlogging > 0) fprintf(stderr,"Error: Cannot open output file %s for BASE64 decoding.",fullMIME_filename);
		//		exit(_EXITERR_BASE64_OUTPUT_NOT_OPEN);
		return -1;
	}



	/* collect prefixing trash (if any, such as spaces, CR's etc)*/
	c = MIME_getchar_start(f);

	/* and push the good char back */
	FFGET_ungetc(c);


	/* do an endless loop, as we're -breaking- out later */
	while (1)
	{

		/* Initialise the decode buffer */
		input[0] = input[1] = input[2] = input[3] = 0;

		/* snatch 4 characters from the input */
		for (i = 0; i < 4; i++)
		{

			/* get a char from the filestream */
			/* This routine was originally a function we called,
			* but due to the fact that this function can get called
			* millions of times, I figured we'd save some CPU by including
			* it "inline" */
			do {
				c = FFGET_fgetc();
			}
			while ( (c != EOF) && ( c < ' ' ) && ( c != '\n' ) && (c != '-') );


			/*	v1.0.0PR5+		c = MIME_getchar(f); -- SEE ABOVE COMMENTS */

			/* if we get a '-', there's a damn good chance we've reached into
			* the next boundary line, which is a rare situation, but it CAN
			* happen, especially if you encode small files ( < 50bytes ) and
			* the MTA doesn't place a \n\r between the line and the next
			* boundary */

			if (c == '-' )
			{
				if (FFGET_fgetc() == '-')
				{
					boundary_crash++;
					eom_reached++;
					break;
				}
			}


			/* if we get a CR then we need to check a few things...*/
			if (c == '\n')
			{

				/* if we've already got our encoded-line length */
				if (p_length != -1)
				{

					/* if the lengths DIFFER, then we have the end of our encoding */
					if (p_length != t_length)
					{
						eom_reached++;
						break;
					} /* p_length != t_length */

				}
				else if (p_length == -1)
				{
					if (t_length > 0) p_length = t_length;
				}



				/* increment the number of -consecutive- CR's we've received... */
				cr_count++;

				/* if we get two consecutive CR's, then we're at the end of the pack */
				if (cr_count > 1)
				{
					eom_reached++;
					break;
				}
				else
				    {
					char_count = 0;
					t_length = 0;
					i--;
					continue;
				} /* else if it wasn't our 2nd CR */

			} /* if c was a \n */
			else
			    {
				cr_count=0;
				t_length++;
			} /* if C == \n */




			/* if we get an EOF char, then we know something went wrong */
			if ( c == EOF )
			{
				if (syslogging > 0) syslog(_SL, "Errror: input stream broken for base64 decoding for file %s\n",hinfo->filename);
				status = -1;
				fclose(of);
				return status;
				break;
			} /* if c was the EOF */




			/* assuming we've gotten this far, then we increment the char_count */
			char_count++;


			/* if we detect the "stopchar" then we better increment the STOP counter */
			if (c == '=')
			{
				stopcount++;
			}

			/* test for and discard invalid chars */
			if (b64[c] == 0x80)
			{
				i--;
				continue;
			}

			/* do the conversion from encoded -> decoded */
			input[i] = (char)b64[c];

		} /* for */


		/* now that our 4-char buffer is full, we can do some fancy bit-shifting and get the required 3-chars of 8-bit data */
		output[0] = (input[0] << 2) | (input[1] >> 4);
		output[1] = (input[1] << 4) | (input[2] >> 2);
		output[2] = (input[2] << 6) | input[3];

		/* determine how many chars to write write and check for errors if our input char count was 4 then we did receive a propper 4:3 Base64 block, hence write it */
		if (i == 4)
		{
			if (fwrite(output, 1, 3-stopcount, of) == EOF)
			{
				/* if there was an error in writing to the file, then we better notify and exit */
				if (syslogging > 0) syslog(_SL,"Error: Cannot write data to %s",fullMIME_filename);
				if (stderrlogging > 0) fprintf(stderr,"Error: Cannot write data to %s",fullMIME_filename);
				//exit(_EXITERR_BASE64_UNABLE_TO_OUTPUT);
				return -1;
			}
		}

		/* tally up our total byte conversion count */
		bytecount+=(3 -stopcount);

		/* if we wrote less than 3 chars, it means we were at the end of the encoded file thus we exit */
		if ((eom_reached)||(stopcount > 0)||(boundary_crash))
		{

			FFGET_closestream();

			/* close the output file, we're done writing to it */
			fclose(of);

			/* if we didn't really write anything, then trash the  file */
			if (bytecount == 0)
			{
				unlink(fullMIME_filename);
			}

			if (boundary_crash) status = 1; // was _BOUNDARY_CRASH

			return status;
		} /* if End-of-MIME or Stopchars appeared */

	} /* while */

	return status;

}







/*----------------------------------------------------------
* MIME_read
*
 * This routine reads in from stdin and saves everything
* to the file mpname, until EOF is receive
*
 * Equivilant under BASH is just 'cat arbfile > mpname'
*
 */
int MIME_read( char *mpname )
{

	char c;
	long int fsize=-1;

	/* open up our input file */
	FILE *fout = fopen(mpname,"w");


	/* check that out file opened up okay */
	if (!fout)
	{
		if (syslogging > 0) syslog(_SL,"Error: Cannot open file %s for writing... check permissions perhaps?",mpname);
		if (stderrlogging > 0) fprintf(stderr,"Error: Cannot open file %s for writing... check permissions perhaps?",mpname);
		//exit(_EXITERR_MIMEREAD_CANNOT_OPEN_OUTPUT);
		return -1;
	}

	/* assuming our file actually opened up */
	if (fout)
	{

		fsize=0;

		/* while there is more data, consume it */
		while ((c = getc(stdin)) != EOF)
		{

			/* write it to file */
			if (fputc(c,fout) != EOF)
			{
				fsize++;
			}
			else
			{
				if (syslogging > 0) syslog(_SL,"Error: Cannot write to file %s... maybe you are out of space?", mpname);
				if (stderrlogging > 0) fprintf(stderr,"Error: Cannot write to file %s... maybe you are out of space?",mpname);
				//exit(_EXITERR_MIMEREAD_CANNOT_WRITE_OUTPUT);
				return -1;
			}
		}

		/* clean up our buffers and close */
		fflush(fout);
		fclose(fout);

	} /* end if fout was received okay */

	/* return our byte count in KB */
	return (int)(fsize /1024);
}



/*---------------------------------------------------
* MIME_init_hexconv()
*
 * Initialises the hex->dec conversion
*
 */
int MIME_init_hexconv( void )
{

	hexconv['0'] = 0;
	hexconv['1'] = 1;
	hexconv['2'] = 2;
	hexconv['3'] = 3;
	hexconv['4'] = 4;
	hexconv['5'] = 5;
	hexconv['6'] = 6;
	hexconv['7'] = 7;
	hexconv['8'] = 8;
	hexconv['9'] = 9;
	hexconv['a'] = 10;
	hexconv['b'] = 11;
	hexconv['c'] = 12;
	hexconv['d'] = 13;
	hexconv['e'] = 14;
	hexconv['f'] = 15;
	hexconv['A'] = 10;
	hexconv['B'] = 11;
	hexconv['C'] = 12;
	hexconv['D'] = 13;
	hexconv['E'] = 14;
	hexconv['F'] = 15;

	return 0;
}


/*----------------------------------------------------------
* MIME_init
*
 * Initialises various aspects required for MIME decoding*/
int MIME_init( void )
{

	/* prepare the base64 conversion dictionary */
	MIME_base64_init();
	MIME_init_hexconv();
	_attachment_count = 0;
	_current_line = 0;
	return 0;
}




/*------------------------------------------------------------------------
Procedure:     MIME_decode_encoding ID:1
Purpose:       Based on the contents of hinfo, this function will call the
required function needed to decode the contents of the file
which is contained within the MIME structure
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIME_decode_encoding( FILE *f, char *unpackdir, struct _header_info *hinfo )
{
	int keep = 1;
	int result = -1;

	// If we have a valid filename, then put it through the process of
	// 	cleaning and filtering
	//
	if (isprint(hinfo->filename[0]))
	{
		MIME_clean_MIME_filename(hinfo->filename);	/* check out thefilename for ISO filenames */
		quick_clean_filename(hinfo->filename); 	/* cleanup garbage characters */
	}

	// If the filename is NOT valid [doesn't have a printable first char]
	// 	then we must create a new file name for it.
	//
	if (!isprint(hinfo->filename[0]))
	{
		sprintf(hinfo->filename,"%s%d",blankfileprefix,filecount);
		filecount++;
		if (_no_nameless) keep = 0;
	}
	else
	    if (strncmp(hinfo->filename,blankfileprefix,strlen(blankfileprefix)) != 0)
	{
		_attachment_count++;
	}

	// If we are required to have "unique" filenames for everything, rather than
	// 	allowing ripMIME to overwrite stuff, then we put the filename through
	//		its tests here
	//
	if ((_unique_names)&&(keep)) MIME_test_uniquename( unpackdir, hinfo->filename, _rename_method );

	// If the calling program requested verbosity, then indicate that we're decoding
	//		the file here
	//
	if ((keep)&&(_verbosity)) fprintf(stdout,"Decoding %s\n", hinfo->filename);


	// Select the decoding method based on the content transfer encoding
	// 	method which we read from the headers
	//
	switch (hinfo->content_transfer_encoding)
	{
	case _CTRANS_ENCODING_B64:
		result = MIME_decode_64(f, unpackdir, hinfo);
		// If the BASE64 attachment is actually a TNEF encoded [MS, you are a pain!]
		// 	then put the created attachment back through another decoding process
		//
		if (hinfo->content_type == _CTYPE_TNEF)
		{
			MIME_decode_TNEF( f, unpackdir, hinfo, 0 );
		}
		break;
	case _CTRANS_ENCODING_7BIT:
		result = MIME_decode_text(f, unpackdir, hinfo, keep);
		break;
	case _CTRANS_ENCODING_8BIT:
		result = MIME_decode_text(f, unpackdir, hinfo, keep);
		break;
	case _CTRANS_ENCODING_QP:
		result = MIME_decode_text(f, unpackdir, hinfo, keep);
		break;
	case _CTRANS_ENCODING_UNKNOWN:
		result = MIME_decode_text(f, unpackdir, hinfo, keep);
		break;
	case _CTRANS_ENCODING_UNSPECIFIED:
		result = MIME_decode_text(f, unpackdir, hinfo, keep);
		break;
	default:
		syslog(1,"MIME_unpack_stage2(): Unknown encoding %d",hinfo->content_transfer_encoding);
		result = -1;
		break;
	}

	return result;
}




/*------------------------------------------------------------------------
Procedure:     MIME_unpack_stage2 ID:1
Purpose:       This function commenced with the file decoding of the attachments
as required by the MIME structure of the file.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MIME_unpack_stage2( FILE *f, char *unpackdir, struct _header_info *hinfo, int current_recursion_level )
{
	int result = 1;
	struct _header_info *h;
	char *p;

//	fprintf(stderr,"Decode Stage II: \n");

	if (current_recursion_level > _RECURSION_LEVEL_MAX)
	{
		if (syslogging > 0) syslog(_SL,"MIME_unpack_stage2(): Current Recursion level of %d is greater than permitted %d",current_recursion_level, _RECURSION_LEVEL_MAX);
		if (stderrlogging > 0) fprintf(stderr,"MIME_unpack_stage2(): Current Recursion level of %d is greater than permitted %d\n",current_recursion_level, _RECURSION_LEVEL_MAX);
		return -1;
	}

	h = hinfo;

	// Get our headers and determin what we have...
	//
//	fprintf(stderr,"Parsing headers (initial)\n");

	result = MIMEH_parse_headers(f,h);

	if (result == -1) return result;

	// If we got a FROM line (mailbox) then exit back out
	//
	if ((_mailbox_format)&&(result == _MIMEH_FOUND_FROM))
	{
		return _MIMEH_FOUND_FROM;
	}



	// If we located a boundary being specified (as apposed to existing)
	// then we push it to the BoundaryStack
	//
	if (h->boundary_located)
	{
//		fprintf(stderr,"Pushing boundary %s\n",h->boundary);
		MIME_BS_push(h->boundary);
		h->boundary_located = 0;
	}
	else
	    {
//		fprintf(stderr,"BOUNDARY-LESS MODE...\n");
		if ((h->content_type == _CTYPE_RFC822)||(h->content_type == _CTYPE_MULTIPART))
		{
			// If there is no filename, then we have a "standard"
			// embedded message, which can be just read off as a
			// continuous stream (simply with new boundaries
			if (( h->content_transfer_encoding != _CTRANS_ENCODING_B64)&&( h->filename[0] == '\0' ))
			{
				h->boundary_located = 0;
				result = MIME_unpack_stage2(f, unpackdir, h, current_recursion_level +1);

				p = MIME_BS_top();
				if (p) strncpy(h->boundary, p,sizeof(h->boundary)-1);
			}
			else
			    {
				result = MIME_decode_encoding( f, unpackdir, h );
				result = MIME_unpack( unpackdir, h->filename, current_recursion_level +1);
			}
		}
		else
		{
//			fprintf(stdout,"Decoding boundaryless file...");
			result = MIME_decode_encoding( f, unpackdir, h );
//			fprintf(stdout,"Result = %d\n",result);
		}

		return result;
	}

	if ((MIME_BS_top()!=NULL)&&(result == 1))
	{

	//	fprintf(stderr,"BOUNDARY LINED decode...\n");

		// Explain this..... if the first headers of the message says RFC822, then we
		// have in effect a wrapped up message, a little crazy to have, but, happens
		// 	So, we dont want to try and read the headers in that case

		result = MIME_decode_encoding(f, unpackdir, h);

		if (result == 1)
		{

			if (MIME_BS_top()!=NULL)
			{

				// As this is a multipart email, then, each section will have its
				// own headers, so, we just simply call the MIMEH_parse call again
				// and get the attachment details

				while (result == 1)
				{
					h->content_type = -1;
					h->filename[0] = '\0';
					h->name[0]     = '\0';
					h->content_transfer_encoding = -1;
					h->content_disposition = -1;

					result = MIMEH_parse_headers(f,h);
					if (h->boundary_located)
					{
//						fprintf(stderr,"2nd-Pushing boundary %s\n",h->boundary);
						MIME_BS_push(h->boundary);
						h->boundary_located = 0;
					}

					if (result == _MIMEH_FOUND_FROM)
					{
						return _MIMEH_FOUND_FROM;
					}

					if (result == 1)
					{

						// If we locate a new boundary specified, it means we have a
						// embedded message, also if we have a ctype of RFC822

						if ( (h->boundary_located) \
						|| (h->content_type == _CTYPE_RFC822)\
					  	|| (h->content_type == _CTYPE_MULTIPART)\
					   )
						{
//							fprintf(stderr,"MULTIPART - DECODE\n");
							/* If there is no filename, then we have a "standard"
							* embedded message, which can be just read off as a
							* continuous stream (simply with new boundaries */
							if (( h->content_transfer_encoding != _CTRANS_ENCODING_B64)&&(h->filename[0] == '\0' ))
							{
//								fprintf(stderr,"NON-BASE64 DECODE\n");
								h->boundary_located = 0;
								result = MIME_unpack_stage2(f, unpackdir, h, current_recursion_level +1);

								p = MIME_BS_top();
								if (p) sprintf(h->boundary,"%s",p);
							}
							else
							    {
								result = MIME_decode_encoding( f, unpackdir, h );
								result = MIME_unpack( unpackdir, h->filename, current_recursion_level +1);
							}
						}
						else
						    {
//							fprintf(stderr,"BASE64 - DECODE\n");
							result = MIME_decode_encoding( f, unpackdir, h );
						}
					}
					else break;

				} // While (result)

				if (result == 1) MIME_BS_pop();

			} // if MIME_BS_top()

		} // if result == 1

	} // if (result)


	return result;
}




/*---------------------------------------------------
* MIME_unpack
*
 * Breaks the MIME package into its various files
*
 * for now we'll use metamail...
*
 */
int MIME_unpack( char *unpackdir, char *mpname, int current_recursion_level )
{

	struct _header_info h;
	int result = 0;
	int status = 0;		/* Global status */
	FILE *f;			/* Pointer for the MIME file we're going to be going through */
	FILE *hf = NULL;

	if (current_recursion_level > _RECURSION_LEVEL_MAX)
	{
		if (syslogging > 0) syslog(_SL,"MIME_unpack_stage2(): Current Recursion level of %d is greater than permitted %d",current_recursion_level, _RECURSION_LEVEL_MAX);
		if (stderrlogging > 0) fprintf(stderr,"MIME_unpack_stage2(): Current Recursion level of %d is greater than permitted %d\n",current_recursion_level, _RECURSION_LEVEL_MAX);
		return -1;
	}

	_current_line = 0;

	/* if we're reading in from STDIN */
	if( mpname[0] == '-' && mpname[1] == '\0' )
	{
		f = stdin;
	}
	else
	    {
		f = fopen(mpname,"r");
		if (!f) syslog(_SL,"MIME_unpack: %s",strerror(errno));
	}

	if ((!hf)&&(_dump_headers))
	{
		hf = fopen(headersname,"w");
		if (!hf)
		{
			_dump_headers = 0;
			if (syslogging > 0) syslog(_SL,"MIME_unpack: Cannot open %s for writing: %s",headersname,strerror(errno));
			if (stderrlogging > 0) fprintf(stderr,"MIME_unpack: Cannot open %s for writing: %s\n",headersname,strerror(errno));
		}
		else
		    {
			MIMEH_set_headers_save(hf);
		}
	}

	/* check to see if we had problems opening the file */
	if (!f)
	{
		if (syslogging > 0) syslog(_SL,"Error: MIME_unpack(): could not open mailpack file \"%s\" : %s",mpname, strerror(errno));
		if (stderrlogging > 0) fprintf(stderr,"Error: MIME_unpack(): could not open mailpack file \"%s\" : %s\n\n",mpname, strerror(errno));
		//exit(_EXITERR_MIMEUNPACK_CANNOT_OPEN_INPUT_FILE);
		return 0; // was -1, but, it causes problem in the mailspool decode (early exits)
//		return -1;
	}

	result = _MIMEH_FOUND_FROM;

	while (result != -1)
	{
		h.content_type = -1;
		h.boundary[0] = '\0';
		h.boundary_located = 0;
		h.filename[0] = '\0';
		h.name[0]     = '\0';
		h.content_transfer_encoding = -1;
		h.content_disposition = -1;
		result = MIME_unpack_stage2(f, unpackdir, &h, 0);
	}


	fclose(f);

	if (_dump_headers)
	{
		if (hf) fclose(hf);
	}

	return status;
}








/*--------------------------------------------------------------------
* MIME_close
*
* Closes the files used in MIME_unpack, such as headers etc */
int MIME_close( void )
{
	if (headers)
	{
		fclose(headers);
	}

	return 0;
}




/*----------END OF MIME.c------------*/

